#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading

from my.instapaper import pages


class Instapaper(Mirror):
    def get_items(self) -> Mirror.Results:
        for page in pages():
            yield node(
                heading=dt_heading(
                    page.dt,
                    f'{link(title="x", url=page.bookmark.instapaper_link)}   {link(title=page.title, url=page.url)}',
                ),
                children=[node(
                    heading=dt_heading(hl.dt, link(title="x", url=page.bookmark.instapaper_link)),
                    body=hl.text,
                    children=[] if hl.note is None else [
                        node(heading=hl.note),
                    ],
                ) for hl in page.highlights]
            )
        # TODO autostrip could be an option for formatter
        # TODO reverse order? not sure...
        # TODO spacing top level items could be option of dumper as well?
        # TODO better error handling, cooperate with org_tools

# TODO move tests to separate files, otherwise they would annoy other people
test = Instapaper.make_test(
    heading='Life Extension Methods',
    contains='sleep a lot',
)


if __name__ == '__main__':
    Instapaper.main()



#!/usr/bin/env python3
"""
Demo Orger adapter for Pocket data. For documentation purposes, so please modify pocket.py if you want to contribute.
"""

"""
First we define some abstractions for Pocket entities (articles and highlights).

While it's not that necessary and for one script you can get away with using json directly,
 it does help to separate parsing and rendering, allows you to reuse parsing for other projects
 and generally makes everything clean.

Also see https://github.com/karlicoss/HPI package for some inspiration.
"""


from datetime import datetime
from pathlib import Path
from typing import NamedTuple, Sequence, Any

class Highlight(NamedTuple):
    """
    Abstract representation of Pocket highlight
    """
    json: Any

    @property
    def text(self) -> str:
        return self.json['quote']

    @property
    def created(self) -> datetime:
        return datetime.strptime(self.json['created_at'], '%Y-%m-%d %H:%M:%S')


class Article(NamedTuple):
    """
    Abstract representation of Pocket saved page
    """
    json: Any

    @property
    def url(self) -> str:
        return self.json['given_url']

    @property
    def title(self) -> str:
        return self.json['given_title']

    @property
    def pocket_link(self) -> str:
        return 'https://app.getpocket.com/read/' + self.json['item_id']

    @property
    def added(self) -> datetime:
        return datetime.fromtimestamp(int(self.json['time_added']))

    @property
    def highlights(self) -> Sequence[Highlight]:
        raw = self.json.get('annotations', [])
        return list(map(Highlight, raw))

    # TODO add tags?


def get_articles(json_path: Path) -> Sequence[Article]:
    """
    Parses Pocket export produced by https://github.com/karlicoss/pockexport
    """
    import json
    raw = json.loads(json_path.read_text())['list']
    return list(map(Article, raw.values()))

"""
Ok, now we can get to implementing the adapter.
"""
from orger import Mirror
"""
Mirror means it's meant to be read-only view onto data (as opposed to Queue).
"""
from orger.inorganic import node, link
from orger.common import dt_heading


class Pocket(Mirror):
    def get_items(self):
        """
        get_items returns a sequence/iterator of nodes
        see orger.inorganic.OrgNode to find out about attributes you can use
        """
        export_file = self.cmdline_args.file # see setup_parser
        for a in get_articles(export_file):
            yield node(
                heading=dt_heading(
                    a.added,
                    # 'pocket' permalink is pretty convenient to jump straight into Pocket app
                    link(title='pocket', url=a.pocket_link)  + ' · ' + link(title=a.title, url=a.url),
                ),
                children=[node( # comments are displayed as org-mode child entries
                    heading=dt_heading(hl.created, hl.text)
                ) for hl in a.highlights]
            )


def setup_parser(p):
    """
    Optional hooks for extra arguments if you need them in your adapter
    """
    p.add_argument('--file', type=Path, help='JSON file from API export', required=True)


if __name__ == '__main__':
    """
    Usage example: ./pocket.py --file /backups/pocket/last-backup.json --to /data/orger/pocket.org
    """
    Pocket.main(setup_parser=setup_parser)

"""
Example pocket.org output:

# AUTOGENERATED BY /L/zzz_syncthing/coding/orger/pocket.py

* [2018-07-09 Mon 10:56] [[https://app.getpocket.com/read/1949330650][pocket]] · [[https://www.gwern.net/Complexity-vs-AI][Complexity no Bar to AI - Gwern.net]]
** [2019-09-22 Sun 03:36] iving up determinism and using randomized algorithms which are faster but may not return an answer or a correct answer1
** [2019-06-22 Sat 16:48] The apparent barrier of a complex problem can be bypassed by (in
* [2016-10-21 Fri 14:42] [[https://app.getpocket.com/read/1407671000][pocket]] · [[https://johncarlosbaez.wordpress.com/2016/09/09/struggles-with-the-continuum-part-2/][Struggles with the Continuum (Part 2) | Azimuth]]
* [2016-05-31 Tue 18:25] [[https://app.getpocket.com/read/1042711293][pocket]] · [[http://www.scottaaronson.com/blog/?p=2464][Bell inequality violation finally done right]]
* [2016-05-31 Tue 18:24] [[https://app.getpocket.com/read/1188624587][pocket]] · [[https://packetzoom.com/blog/how-to-test-your-app-in-different-network-conditions.html][How to test your app in different network conditions -]]
* [2016-05-31 Tue 18:24] [[https://app.getpocket.com/read/1191143185][pocket]] · [[http://www.schibsted.pl/2016/02/hood-okhttps-cache/][What's under the hood of the OkHttp's cache?]]
* [2016-03-15 Tue 17:27] [[https://app.getpocket.com/read/1187239791][pocket]] · [[http://joeduffyblog.com/2016/02/07/the-error-model/][Joe Duffy - The Error Model]]
** [2019-09-25 Wed 18:20] A bug is a kind of error the programmer didn’t expect. Inputs weren’t validated correctly, logic was written wrong, or any host of problems have arisen.
** [2019-09-25 Wed 18:19] First, throwing an exception is usually ridiculously expensive. This is almost always due to the gathering of a stack trace.
** [2019-09-25 Wed 18:20] In other words, an exception, as with error codes, is just a different kind of return value!
"""



from datetime import datetime, date
import logging
from pathlib import Path
import re
import os
from collections import OrderedDict
from typing import NamedTuple, Optional, Sequence, Dict, Mapping, Any, Tuple, TypeVar, Callable, Union, List, TYPE_CHECKING
import warnings

# TODO settings object? not ideal...

# todo use mypy literals later?
from enum import Enum
class TimestampStyle(Enum):
    INACTIVE = ('[', ']')
    ACTIVE   = ('<', '>')
    PLAIN    = ('' , '')
    NONE     = ()


Dateish = Union[datetime, date]

PathIsh = Union[str, Path]

def link(*, url: PathIsh, title: Optional[str]) -> str:
    """
    >>> link(url='http://reddit.com', title='[R]eddit!')
    '[[http://reddit.com][Reddit!]]'
    """
    U = _sanitize_url(str(url))
    if title == '':
        # org mode doesn't like empty titles..
        # TODO sanitize_title?
        title = None
    if title is not None:
        title = _sanitize_heading(title)
        return f'[[{U}][{title}]]'
    else:
        return f'[[{U}]]'



def timestamp(dt: Dateish, inactive: bool=False, active: bool=False) -> str:
    """
    default is active
    >>> dt = datetime.strptime('19920110 04:45', '%Y%m%d %H:%M')
    >>> timestamp(dt)
    '<1992-01-10 Fri 04:45>'
    >>> timestamp(dt, inactive=True)
    '[1992-01-10 Fri 04:45]'
    """
    style: TimestampStyle
    if inactive and active:
        warnings.warn(f"Please specify one of 'inactive' or 'active' for {dt}")
        # arbitrary choice: we let active win. So the uses sees the offending entry on agenda
        style = TimestampStyle.ACTIVE
    elif inactive:
        style = TimestampStyle.INACTIVE
    else: # active
        style = TimestampStyle.ACTIVE
    return timestamp_with_style(dt=dt, style=style)


def timestamp_with_style(dt: Dateish, style: TimestampStyle) -> str:
    """
    >>> dt = datetime.max
    >>> timestamp_with_style(dt, TimestampStyle.NONE)
    ''
    >>> timestamp_with_style(dt, TimestampStyle.PLAIN)
    '9999-12-31 Fri 23:59'
    """
    if style is TimestampStyle.NONE:
        return ''
    beg, end = style.value
    r = asorgdate(dt)
    if isinstance(dt, datetime):
        r += " " + asorgtime(dt)
    return beg + r + end


# TODO priority
# TODO for sanitizing, have two strategies: error and replace?
def asorgoutline(
        heading: Optional[str] = None,
        todo: Optional[str] = None,
        tags: Sequence[str] = [],
        scheduled: Optional[Dateish] = None,
        # TODO perhaps allow list of tuples? properties might be repeating
        properties: Optional[Mapping[str, str]]=None,
        body: Optional[str] = None,
        level: int=1,
        escaped: bool=False,
) -> str:
    r"""
    Renders Org mode outline (apart from children)

    >>> asorgoutline(
    ...     heading=None,
    ...     tags=['hi'],
    ...     body='whatever...'
    ... )
    '* :hi:\n whatever...'
    >>> asorgoutline(heading=None, todo=None, tags=(), level=2)
    '** '
    >>> asorgoutline(heading='heading', body=None)
    '* heading'
    >>> asorgoutline(heading='heading', body='keep\n newlines\n')
    '* heading\n keep\n  newlines\n'
    >>> asorgoutline(heading='123', todo='TODO', level=0)
    'TODO 123'
    >>> asorgoutline(heading='*abacaba', body='***whoops', tags=('baa@d tag', 'goodtag'))
    '* *abacaba :baa@d_tag:goodtag:\n ***whoops'
    >>> asorgoutline(heading='just heading', level=0)
    'just heading'
    >>> asorgoutline(heading='', level=0)
    ''
    >>> asorgoutline(heading='task', body='hello', scheduled=datetime.utcfromtimestamp(0))
    '* task\nSCHEDULED: <1970-01-01 Thu 00:00>\n hello'
    """
    if heading is None:
        heading = ''
    # TODO reuse sanitizing?
    if not escaped:
        heading = re.sub(r'\s', ' ', heading)

    # TODO not great that we always pad body I guess. maybe needs some sort of raw_body argument?
    if body is not None and not escaped:
        body = _sanitize_body(body)

    parts = []

    if level > 0:
        parts.append('*' * level)

    if todo is not None:
        parts.append(todo)

    if len(heading) > 0:
        parts.append(heading)

    if len(tags) > 0:
        # tags_s = ('' if heading.endswith(' ') else ' ') +
        tags_s = ':' + ':'.join(map(_sanitize_tag, tags)) + ':'
        parts.append(tags_s)

    sch_lines = [] if scheduled is None else [
        'SCHEDULED: ' + timestamp(scheduled, active=True)
    ]

    props_lines: List[str] = []
    props = {} if properties is None else properties
    if len(props) > 0:
        props_lines.append(':PROPERTIES:')
        props_lines.extend(f':{prop}: {value}' for prop, value in props.items())
        props_lines.append(':END:')

    body_lines = [] if body is None else [body]

    if level > 0 and len(parts) == 1:
        # means it's only got level stars, so we need to make sure space is present (otherwise it's not an outline)
        parts.append('')
    lines = [
        ' '.join(parts), # TODO just in case check that parts doesn't have newlines?
        *sch_lines,
        *props_lines,
        *body_lines,
    ]
    # TODO careful here, I guess actually need some tests for endlines
    return '\n'.join(lines)


T = TypeVar('T')
Lazy = Union[T, Callable[[], T]]


# jeez...
# otherwise it struggles to resolve recursive type
# could as well make it dataclass... just dunno if should introduce dependency for py36
if TYPE_CHECKING:
    from dataclasses import dataclass as mypy_dataclass
    Base = object
else:
    mypy_dataclass = lambda x: x
    Base = NamedTuple


@mypy_dataclass
class OrgNode(Base):
    """
    Meant to be somewhat compatible with https://orgparse.readthedocs.io/en/latest/#orgparse.node.OrgNode
    """
    heading: Lazy[str]
    todo: Optional[str] = None
    tags: Sequence[str] = ()
    scheduled: Optional[Dateish] = None
    properties: Optional[Mapping[str, str]] = None
    # TODO make body lazy as well?
    body: Optional[str] = None
    children: Sequence['OrgNode'] = ()

    # NOTE: this is a 'private' attribute
    escaped: bool=False

    def _render_self(self) -> str:
        return asorgoutline(
            heading=_from_lazy(self.heading),
            todo=self.todo,
            tags=self.tags,
            properties=self.properties,
            scheduled=self.scheduled,
            body=self.body,
            level=0,
            escaped=self.escaped,
        )

    def _render_hier(self) -> List[Tuple[int, str]]:
        res = [(0, self._render_self())]
        for ch in self.children:
            # TODO make sure there is a space??
            # TODO shit, would be nice to tabulate?.. not sure
            res.extend((l + 1, x) for l, x in ch._render_hier())
        return res

    def render(self, level: int=1) -> str:
        r"""
        >>> OrgNode('something', todo='TODO').render()
        '* TODO something'
        >>> OrgNode('something else').render()
        '* something else'
        >>> OrgNode(heading=lambda: 'hi', body='so lazy...').render()
        '* hi\n so lazy...'
        >>> OrgNode('#+FILETAGS: sometag', children=[OrgNode('subitem')]).render(level=0)
        '#+FILETAGS: sometag\n* subitem'
        """
        rh = self._render_hier()
        rh = [(level + l, x) for l, x in rh]
        return '\n'.join('*' * l + (' ' if l > 0 else '') + x for l, x in rh)

node = OrgNode


## helper functions

def asorgdate(t: Dateish) -> str:
    return t.strftime("%Y-%m-%d %a")


def asorgtime(t: datetime) -> str:
    return t.strftime("%H:%M")


# meh
def _from_lazy(x: Lazy[T]) -> T:
    if callable(x):
        return x()
    else:
        return x


from typing import Mapping
def maketrans(d: Dict[str, str]) -> Dict[int, str]:
    # make mypy happy... https://github.com/python/mypy/issues/4374
    return str.maketrans(d)


def _sanitize_url(x: str) -> str:
    r'''
    Sanitize url to put into [] safely
    >>> _sanitize_url("/path/to/[Baez] Lectures on Classical Mechanics.pdf")
    '/path/to/\\[Baez\\] Lectures on Classical Mechanics.pdf'
    '''
    # might be incomplete..
    return x.translate(maketrans({
        '[': r'\[',
        ']': r'\]',
    }))


def _sanitize_heading(x: str) -> str:
    # TODO do something smarter? e.g. https://stackoverflow.com/questions/12737564/escaping-characters-in-emacs-org-mode
    return re.sub(r'[\]\[]', '', x)


# TODO allow passing raw body?
def _sanitize_body(text: str) -> str:
    r"""
    >>> _sanitize_body('this is not a heading!:\n* hi')
    ' this is not a heading!:\n * hi'
    >>> _sanitize_body('Some thoughts:\r\n\r\n* convenience')
    ' Some thoughts:\n \n * convenience'
    """
    # TODO hmm. maybe just tabulating with 1 space is enough?...
    text = text.replace('\r\n', os.linesep)
    return ''.join(' ' + l for l in text.splitlines(keepends=True))


def _sanitize_tag(tag: str) -> str:
    """
    >>> _sanitize_tag('test-d@shes')
    'test_d@shes'
    """
    # https://orgmode.org/manual/Tags.html
    # Tags are normal words containing letters, numbers, ‘_’, and ‘@’.
    # TODO not sure, perhaps we want strict mode for formatting?
    # TODO reuse orgparse regexes?
    return re.sub(r'[^@\w]', '_', tag)




#!/usr/bin/env python3
from orger import InteractiveView
from orger.inorganic import node, link
from orger.common import todo

from my.books.kobo import get_todos, Highlight # type: ignore


class KoboTodos(InteractiveView):
    def get_items(self):
        for t in get_todos():
            # TODO shit judging by the state.json, looks like eid might be flaky?
            yield t.eid, todo(
                t.dt,

                heading=t.text,
                tags=['kobo2org'],
                body=f'{t.annotation}\nfrom {t.book}\n',
            )

# TODO test?

if __name__ == '__main__':
    KoboTodos.main()



#!/usr/bin/env python3
"""
Sometimes when I'm annotating using [[https://hypothes.is][Hypothesis]], I want to think more
about specific highlights, google more about them later or generally act on them somehow.

Normally you'd have to copy the URL, highlighted text and create a task from it.

This script does that automatically, only thing that you have to do is to mark it with a tag or type 'todo'
in the annotation text.

Items get scheduled and appear on my org-mode agenda,
so I can un/reschedule them if they don't require immediate attention.
"""

from orger import InteractiveView
from orger.inorganic import node, link
from orger.common import todo

from my.hypothesis import get_highlights, Highlight


def is_todo(e: Highlight) -> bool:
    if any(t.lower() == 'todo' for t in e.tags):
        return True
    text = e.annotation or ''
    return text.lstrip().lower().startswith('todo')


class HypTodos(InteractiveView):
    def get_items(self):
        for t in get_highlights():
            if isinstance(t, Exception):
                # I guess there isn't much we can do here? will be spotted by other tools
                continue
            if not is_todo(t):
                continue
            yield t.hid, todo(
                dt=t.created,

                heading=t.highlight,
                tags=['hyp2org', *t.tags],
                body=f'''
{t.annotation}
{link(title=t.title, url=t.url)}
{link(title="in context", url=t.hyp_link)}
'''.lstrip(),
            )


if __name__ == '__main__':
    HypTodos.main()



#!/usr/bin/env python3
"""
Automatically import stuff from my Kobo backups into org-mode for further spaced repetition (via org-drill in my case).
Mainly learning vocabulary with it.

It uses [[https://github.com/karlicoss/kobuddy][kobuddy]] to extract data from Kobo exports.

The name stands for K[oboD]rill.
"""
from orger import InteractiveView
from orger.inorganic import node, link
from orger.common import todo

from my.books.kobo import get_highlights, Highlight # type: ignore

def is_drill(i: Highlight):
    if i.kind == 'bookmark':
        return False
    ann = i.annotation or ''
    if ann.strip().lower() == 'drill':
        return True

    # single highlighted word almost always means I want to memorise it
    # might result in a false positive but not too much of a problem
    words = (i.text or '').strip().split()
    if len(words) <= 1:
        return True
    return False


def get_drill_items():
    return list(sorted(filter(is_drill, get_highlights()), key=lambda h: h.dt))


class Krill(InteractiveView):
    def get_items(self):
        for i in get_drill_items():
            yield i.eid, todo(
                i.dt,

                heading=i.text,
                tags=['drill'],
                body=f'from {i.book.title}\n',
            )

if __name__ == '__main__':
    Krill.main()



#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading

import my.pinboard as pinboard

class PinboardView(StaticView):
    def get_items(self):
        def make_item(b: pinboard.Bookmark):
            return node(
                heading=dt_heading(b.created, link(title=b.title, url=b.url)),
                body=b.description,
                tags=b.tags,
            )
        return [make_item(b) for b in pinboard.bookmarks()]


test = PinboardView.make_test(
    heading='Cartesian Closed Comic #21',
    contains='doctorwho', # TODO predicate?
)

if __name__ == '__main__':
    PinboardView.main()




#!/usr/bin/env python3
import logging
from typing import Union, Optional

Level = int
LevelIsh = Optional[Union[Level, str]]


def mklevel(level: LevelIsh) -> Level:
    if level is None:
        return logging.NOTSET
    if isinstance(level, int):
        return level
    return getattr(logging, level.upper())


_FMT = '{start}[%(levelname)-7s %(asctime)s %(name)s %(filename)s:%(lineno)d]{end} %(message)s'
_FMT_COLOR   = _FMT.format(start='%(color)s', end='%(end_color)s')
_FMT_NOCOLOR = _FMT.format(start='', end='')


def setup_logger(logger: logging.Logger, level: Level) -> None:
    try:
        import logzero # type: ignore
    except ModuleNotFoundError:
        import warnings
        warnings.warn("You might want to install 'logzero' for nice colored logs!")
        logger.setLevel(level)
        h = logging.StreamHandler()
        h.setLevel(level)
        h.setFormatter(logging.Formatter(fmt=_FMT_NOCOLOR))
        logger.addHandler(h)
    else:
        formatter = logzero.LogFormatter(
            fmt=_FMT_COLOR,
            datefmt=None, # pass None to prevent logzero from messing with date format
        )
        logzero.setup_logger(logger.name, level=level, formatter=formatter)


class LazyLogger(logging.Logger):
    # TODO perhaps should use __new__?

    def __new__(cls, name, level: LevelIsh = 'DEBUG'):
        logger = logging.getLogger(name)
        lvl = mklevel(level)

        # this is called prior to all _log calls so makes sense to do it here?
        def isEnabledFor_lazyinit(*args, logger=logger, orig=logger.isEnabledFor, **kwargs):
            att = 'lazylogger_init_done'
            if not getattr(logger, att, False): # init once, if necessary
                setup_logger(logger, level=lvl)
                setattr(logger, att, True)
            return orig(*args, **kwargs)

        logger.isEnabledFor = isEnabledFor_lazyinit  # type: ignore[assignment]
        return logger


def test():
    ll = LazyLogger('test')
    ll.debug('THIS IS DEBUG')
    ll.warning('THIS IS WARNING')
    ll.exception(RuntimeError("oops"))


if __name__ == '__main__':
    test()



#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading, setup_logger, error

import logging
from datetime import datetime

import my.pdfs as pdfs


class PdfView(StaticView):
    def get_items(self):
        for pdf in sorted(
                pdfs.annotated_pdfs(),
                key=lambda p: datetime.min if isinstance(p, Exception) or p.date is None else p.date.replace(tzinfo=None),
        ):
            if isinstance(pdf, Exception):
                yield error(pdf)
                continue
            yield node(
                dt_heading(pdf.date, str(pdf.path)),
                children=[node(
                    dt_heading(a.date, f'page {a.page} {a.highlight or ""}'),
                    body=a.comment,
                ) for a in pdf.annotations]
            )


if __name__ == '__main__':
    setup_logger(pdfs.get_logger(), level=logging.DEBUG)
    PdfView.main()



from datetime import datetime
from typing import Optional
from pathlib import Path

from .inorganic import OrgNode, timestamp, timestamp_with_style, TimestampStyle


class settings:
    DEFAULT_TIMESTAMP_STYLE = TimestampStyle.INACTIVE
    USE_PANDOC: bool = True


def dt_heading(dt: Optional[datetime], heading: str) -> str:
    """
    Helper to inline datetime in heading
    """
    # TODO move to inorganic? not sure
    if dt is None:
        return heading
    else:
        return timestamp_with_style(dt=dt, style=settings.DEFAULT_TIMESTAMP_STYLE) + ' ' + heading


def error(e: Exception) -> OrgNode:
    import traceback
    return OrgNode(
        heading=f"ERROR!",
        body='\n'.join(traceback.format_exception(Exception, e, None)),
    )


# TODO not sure if belongs here
def todo(dt: datetime, **kwargs):
    """
    Helper to simplify creating todos
    """
    props = kwargs.get('properties', {})
    props.update({'CREATED': timestamp(dt, inactive=True)})
    kwargs['properties'] = props
    return OrgNode(
        todo='TODO',
        scheduled=dt.date(),
        **kwargs,
    )


# todo use klogging2?
from .klogging import LazyLogger, setup_logger


def orger_user_dir() -> Path:
    import appdirs # type: ignore[import]
    return Path(appdirs.user_config_dir('orger'))



from pathlib import Path
from os.path import lexists
import logging
from typing import Union

PathIsh = Union[str, Path]

def atomic_append_raw(
        path: PathIsh,
        data: str,
):
    path = Path(path)
    # https://stackoverflow.com/a/13232181
    enc = data.encode('utf8')
    # TODO handle windows properly? https://www.notthewizard.com/2014/06/17/are-files-appends-really-atomic/
    if len(enc) > 4096:
        logging.warning("writing out %s might be non-atomic (see https://stackoverflow.com/a/1154599/706389)", data)
    with path.open('ab') as fo:
        fo.write(enc)


# TODO might be useful in other projects?
def assert_not_edited(path: Path):
    vim = '.' + path.name + '.swp'
    emacs = '.#' + path.name
    for x in [vim, emacs]:
        lf = path.parent / x
        if lexists(lf): # lexist is necessary because emacs uses symlink for lock file
            raise RuntimeError('File is being edited: {}'.format(lf))


def atomic_append_check(
        path: PathIsh,
        data: str,
):
    """
    This is editor (emacs/vim)-aware and checks for existence of swap file first.
    Not fully atomic, but hopefully atomic enough for all practical purposes
    TODO make it configurabe if user has different swap file patterns for some reason?
    """
    # may be a bit too paranoid, but perhaps doesn't hurt.
    path = Path(path)
    assert_not_edited(path)
    atomic_append_raw(path, data)


def test_atomic_append_check(tmp_path):
    of = tmp_path / 'test.org'
    of.touch()

    import pytest # type: ignore

    from subprocess import Popen, PIPE, check_call
    from time import sleep

    from contextlib import contextmanager
    @contextmanager
    def tmp_popen(*args, **kwargs):
        with Popen(*args, **kwargs) as p:
            try:
                yield p
            finally:
                p.terminate()

    atomic_append_check(of, 'data1')
    atomic_append_check(of, 'data2')
    assert of.read_text() == 'data1data2'

    with tmp_popen(['vi', '-c', 'startinsert', str(of)], stdin=PIPE, stdout=PIPE, stderr=PIPE) as p: # enter insert mode
        for attempt in range(10):
            # ugh, needs long pause for some reason
            sleep(1)
            swapfiles = list(tmp_path.glob('.*.swp'))
            if len(swapfiles) > 0:
                break
        else:
            raise AssertionError(f'Expected swapfiles in {tmp_path}')

        with pytest.raises(Exception):
            # Expected to raise due to being edited by vim
            atomic_append_check(of, 'test')



#!/usr/bin/env python3
from itertools import chain
from typing import Iterable

from orger import Mirror
from orger.inorganic import node, link, OrgNode
from orger.common import dt_heading
from orger import pandoc

import my.roamresearch as roamresearch


# todo ^^ ^^ things are highlight?
def roam_text_to_org(text: str) -> str:
    """
    Cleans up Roam artifacts and adapts for better Org rendering
    """
    for f, t in [
            ('{{[[slider]]}}', ''),
    ]:
        text = text.replace(f, t)
    org = pandoc.to_org(text, from_='markdown')
    org = org.replace(r'\_', '_') # unescape, it's a bit aggressive..
    return org


def roam_note_to_org(node: roamresearch.Node, top=False) -> Iterable[OrgNode]:
    """
    Converts Roam node into Org-mode representation
    """
    children = list(chain.from_iterable(map(roam_note_to_org, node.children)))

    empty = len(node.body or '') == 0 and len(children) == 0
    if empty:
        # sometimes nodes are empty. two cases:
        # - no heading -- child notes, like accidental enter presses I guess
        # - heading    -- notes that haven't been created yet
        # just don't do anything in this case
        # todo make this logic conditional?
        return

    title = node.title
    # org-mode target allows jumping straight into
    # conveniently, links in Roam are already represented as [[link]] !
    target = '' if title is None else f'<<{title}>> '
    heading = target + link(title='x', url=node.permalink)

    todo = None
    body = node.body
    if body is not None:
        for t in ('TODO', 'DONE'):
            ts = '{{[[' + t + ']]}}'
            if body.startswith(ts):
                todo = t
                body = body[len(ts):]

        body = roam_text_to_org(body)

        lines = body.splitlines(keepends=True)
        # display first link of the body as the heading
        if len(lines) > 0:
            heading = heading + ' ' + lines[0]
            body = ''.join(lines[1:])
            if len(body) == 0:
                body = None

    if top:
        heading = dt_heading(node.created, heading)

    yield OrgNode(
        todo=todo,
        heading=heading,
        body=body,
        children=children,
    )


class RoamView(Mirror):
    def get_items(self):
        rr = roamresearch.roam()
        from concurrent.futures import ThreadPoolExecutor
        # todo might be an overkill, only using because of pandoc..
        with ThreadPoolExecutor() as pool:
            items = list(chain.from_iterable(pool.map(roam_note_to_org, rr.notes)))

        # move the ones with no children to the bottom
        items = list(sorted(items, key=lambda n: len(n.children), reverse=True))

        yield from items


if __name__ == '__main__':
    RoamView.main()



#!/usr/bin/env python3
"""
Module for Materialistic app for Hackernews
https://play.google.com/store/apps/details?id=io.github.hidroh.materialistic
"""

from orger import Queue
from orger.inorganic import node, link
from orger.common import dt_heading

from my.materialistic import saves

class Materialistic(Queue):
    def get_items(self):
        for s in saves():
            yield s.uid, node(
                heading=dt_heading(
                    s.when,
                    link(title=s.title, url=s.hackernews_link),
                ),
                body=s.url,
            )


if __name__ == '__main__':
    Materialistic.main()



#!/usr/bin/env python3
from orger import InteractiveView
from orger.inorganic import node, link
from orger.common import todo

from my.instapaper import pages, is_todo


class IpTodos(InteractiveView):
    def get_items(self):
        for page in pages():
            bm = page.bookmark
            for hl in page.highlights:
                if not is_todo(hl):
                    continue
                yield hl.hid, todo(
                    dt=hl.dt,
                    heading=hl.text,
                    tags=['ip2org'],
                    body=f'{hl.note}\nfrom {link(title="x", url=hl.instapaper_link)}   {link(title=bm.title, url=bm.url)}',
                )

if __name__ == '__main__':
    IpTodos.main()


# TODO move error handling to the base class
# TODO add some test?



#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading


from my.media.movies import get_movies

class Movies(StaticView):
    def get_items(self):
        movies = get_movies()
        for m in movies:
            yield m.title, node(
                dt_heading(m.created, m.title),
                body=f'rating: {m.rating}',
            )


if __name__ == '__main__':
    Movies.main()



#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading

from my.rtm import active_tasks


class RtmView(StaticView):
    def get_items(self):
        for t in active_tasks():
            yield t.uid, node(
                dt_heading(t.time, t.title),
                tags=t.tags,
                body='\n'.join(t.notes),
            )

if __name__ == '__main__':
    RtmView.main()



"""
Helper for converting stuff to pandoc
"""
import logging
import shutil
from subprocess import run, PIPE
from typing import Optional


from .common import settings

if settings.USE_PANDOC:
    has_pandoc = shutil.which('pandoc') is not None

    if not has_pandoc:
        import warnings
        warnings.warn("Please install 'pandoc' to convert HTML to org-mode. See https://pandoc.org/installing.html")
        settings.USE_PANDOC = False


def to_org(data: str, *, from_: str, logger=logging) -> str:
    if not settings.USE_PANDOC:
        return data
    # TODO batch??

    # meh. for some reason they are converted to \\ otherwise
    if from_ == 'html':
        data = data.replace('<br>', '')
   
    try:
        r = run(
            ['pandoc', '-f', from_, '-t', 'org', '--wrap=none'],
            check=True,
            input=data.encode('utf8'),
            stdout=PIPE,
        )
    except Exception as e:
        logger.exception(e)
        return data # fallback
    res = r.stdout.decode('utf8')
    return res


def test():
    # html = "<p>and a <i>comment</i> too&nbsp;</p><p><br></p><p><b>multiline</b>!</p>"
    # TODO ok, it's annoying... not sure what to do with nonpritable crap
    html = "<p>and a <i>comment</i> too</p><p><br></p><p><b>multiline</b>!</p>"
    assert to_org(data=html, from_='html') == r'''
and a /comment/ too

*multiline*!
'''.lstrip()



#!/usr/bin/env python3
import argparse
from argparse import ArgumentParser, Namespace
import logging
import inspect
from pathlib import Path
from subprocess import check_call
from tempfile import TemporaryDirectory
from typing import List, Tuple, Iterable, Optional, Union

from .inorganic import OrgNode, TimestampStyle
from .state import JsonState
from .atomic_append import PathIsh, atomic_append_check, assert_not_edited
from .common import setup_logger, orger_user_dir

# TODO tests for determinism? not sure where should they be...
# think of some generic thing to test that?

Key = str
OrgWithKey = Tuple[Key, OrgNode]


_style_map = {
    k.lower(): v
    for k, v in TimestampStyle._member_map_.items() # type: ignore[attr-defined]
}

class OrgView:
    logger_tag: Optional[str] = None
    DEFAULT_HEADER: str = '# should be overridden'

    # TODO cmdline args shouldn't be none?
    def __init__(
            self,
            cmdline_args: Optional[Namespace]=None,
            file_header: Optional[str]=None,
    ) -> None:
        self.cmdline_args: Namespace = cmdline_args if cmdline_args is not None else Namespace()
        tag = self.name() if self.logger_tag is None else self.logger_tag
        self.logger = logging.getLogger(tag)

        tool = Path(inspect.getfile(self.__class__)).absolute()
        self.file_header = file_header if file_header is not None else self.DEFAULT_HEADER.format(tool=tool)

    @property
    def args(self) -> Namespace:
        # TODO deprecate cmdline_args?
        return self.cmdline_args

    @classmethod
    def name(cls):
        return cls.__name__

    def get_items(self) -> Iterable:
        raise NotImplementedError

    def main_common(self) -> None:
        timestamp_style = self.args.timestamps
        from .common import settings
        # hacky, but does the trick for now...
        settings.DEFAULT_TIMESTAMP_STYLE = _style_map[timestamp_style]
        setup_logger(self.logger, level=logging.DEBUG)

        pandoc = self.args.pandoc
        settings.USE_PANDOC = pandoc

    @classmethod
    def parser(cls) -> ArgumentParser:
        F = lambda prog: argparse.ArgumentDefaultsHelpFormatter(prog, width=120)
        p = argparse.ArgumentParser(formatter_class=F) # type: ignore

        p.add_argument(
            '--disable-pandoc',
            action='store_false',
            dest='pandoc',
            help='Pass to disable pandoc conversions to org-mode (it might be slow in some cases)',
        )
        p.add_argument(
            '--timestamps',
            type=str,
            choices=list(_style_map.keys()),
            default='inactive',
            help="timestamp style, default '%(default)s'",
        )
        return p


# TODO wonder if I could reuse append bits here?
class Mirror(OrgView):
    """
    *Mirror* (old name =StaticView=): mirrors *all data* from a source, and generated from scratch every time, hence *read only*.
    """

    DEFAULT_HEADER = '''
# This file is AUTOGENERATED by {tool}
# It's deliberately read-only, because it will be overwritten next time Orger is run.
# If you want to edit it anyway, you can use chmod +w in your terminal, or M-x toggle-read-only in Emacs.
'''.lstrip()

    # allowed to be either for Mirror
    Results = Iterable[Union[OrgNode, OrgWithKey]]

    @classmethod
    def main(cls, setup_parser=None) -> None:
        p = cls.parser()
        p.add_argument('--to', type=Path, default=Path(cls.name() + '.org'), help='Filename to output')
        if setup_parser is not None:
            setup_parser(p)

        args = p.parse_args()
        inst = cls(cmdline_args=args)
        inst.main_common()
        inst._run(to=args.to)

    def get_items(self) -> Iterable:
        raise NotImplementedError

    def _run(self, to: Path):
        org_tree = self.make_tree()
        rtree = org_tree.render(level=0)

        assert_not_edited(to)
        # again, not properly atomic, but hopefully enough
        # TODO create a github issue, maybe someone comes up with proper way of solving this
        to.touch()
        check_call(['chmod', '+w', to])
        to.write_text(rtree)
        check_call(['chmod', '-w', to])


    def make_tree(self) -> OrgNode:
        items: List[OrgNode] = []
        for p in self.get_items():
            # it's ok to use items without keys in View
            if isinstance(p, OrgNode):
                items.append(p)
            else:
                items.append(p[1]) # presumably, OrgWithKey

        split = self.file_header.splitlines(keepends=True)
        heading = split[0].rstrip()
        body = ''.join(split[1:])
        return OrgNode(
            # TODO shit. are newlines sanitized from file header??
            heading=heading,
            body=body,
            children=items,
            escaped=True,
        )

    @classmethod
    def make_test(cls, *, heading: str, contains: Optional[str]=None):
        from .inorganic import _from_lazy
        def pick_heading(root: OrgNode, text: str) -> Optional[OrgNode]:
            if text in _from_lazy(root.heading):
                return root
            for ch in root.children:
                chr = pick_heading(ch, text)
                if chr is not None:
                    return chr
            else:
                return None

        def test():
            tree = cls().make_tree() # TODO make sure it works on both static and interactive?
            ll = pick_heading(tree, heading)
            assert ll is not None
            if contains is not None:
                assert contains in ll.render()
        return test
StaticView = Mirror


class Queue(OrgView):
    """
    *Queue* (old name =InteractiveView=): works as a queue, *only previously unseen items* from the data source are added to the output org-mode file.

    To keep track of previously seen iteems, it's using a separate JSON =state= file.

    A typical usecase is a todo list, or a content processing queue.
    You can use such a module as you use any other org-mode file: schedule/refile/comment/set priorities, etc.
    """

    DEFAULT_HEADER = '''
# This file is AUTOGENERATED by {tool}
'''.lstrip()

    def _run(
            self,
            to: Path,
            state_path: Path,
            init: bool=False,
            dry_run: bool=False,
    ) -> None:
        if not to.exists() and not init:
            err = RuntimeError(f"{to} doesn't exist! Try running with --init")
            import sys
            if sys.stdin.isatty():
                resp = input(f"{to} doesn't exist. Create empty file? y/n ").strip().lower()
                if resp != 'y':
                    raise err
            else:
                raise err

        state_path.parent.mkdir(parents=True, exist_ok=True) # not sure...
        state = JsonState(
            path=state_path,
            logger=self.logger,
            dry_run=dry_run,
        )
        items = list(self.get_items())

        from collections import Counter
        dups = [k for k, cnt in Counter(i[0] for i in items).items() if cnt > 1]
        if len(dups) > 0:
            raise RuntimeError(f'Duplicate items {dups}')

        if not to.exists():
            self.logger.warning("target %s didn't exist, initializing!", to)
            atomic_append_check(to, self.file_header + '\n')

        for key, item in items:
            def action(item=item):
                # not sure about this newline, but better to have extra whitespace than rely on trailing
                rendered = '\n' + item.render(level=1)
                atomic_append_check(
                    to,
                    rendered,
                )
            self.logger.debug('processing %s', key)
            state.feed(
                key=key,
                value=item, # TODO not sure about this one... perhaps only link?
                action=action,
            )

    def get_items(self) -> Iterable[OrgWithKey]:
        raise NotImplementedError

    @classmethod
    def main(cls, setup_parser=None) -> None:
        default_state = orger_user_dir() / 'states' / (cls.name() + '.state.json')
        p = cls.parser()
        p.add_argument('--to'   , type=Path, default=Path(cls.name() + '.org')       , help='file where new items are added')
        p.add_argument('--state', type=Path, default=default_state, help='state file for keeping track of handled items')
        p.add_argument('--init', action='store_true') # todo not sure if I really need it?
        p.add_argument('--dry-run', action='store_true', help='Run without modifying the state file')
        if setup_parser is not None:
            setup_parser(p)

        args = p.parse_args()
        inst = cls(cmdline_args=args)
        inst.main_common()
        inst._run(
            to=args.to,
            state_path=args.state,
            init=args.init,
            dry_run=args.dry_run,
        )
InteractiveView = Queue


def test_org_view_overwrite(tmp_path: Path):
    class TestView(StaticView):
        def __init__(self, items: List[OrgWithKey], *args, **kwargs) -> None:
            super().__init__(*args, file_header='# autogenerated!\n#+TITLE: sometitle\nalso text\n', **kwargs) # type: ignore
            self.items = items

        def get_items(self):
            return self.items

    rpath = tmp_path / 'test.org'

    TestView([])._run(to=rpath)
    assert rpath.read_text() == '''
# autogenerated!
#+TITLE: sometitle
also text
'''.lstrip()

    TestView([
        # TODO shit, it's gonna use implicit date??
        ('first' , OrgNode(heading='whatever')),
        ('second', OrgNode(heading='alala')), # TODO why was that even necessary??
    ])._run(to=rpath)
    # TODO eh, perhaps use trailing space?
    assert rpath.read_text() == """
# autogenerated!
#+TITLE: sometitle
also text

* whatever
* alala""".lstrip()


def test_org_view_append(tmp_path: Path):
    import json
    class TestView(InteractiveView):
        def __init__(self, items: List[OrgWithKey], *args, **kwargs) -> None:
            super().__init__(*args, file_header='# autogenerated!', **kwargs) # type: ignore
            self.items = items

        def get_items(self):
            for i in self.items:
                yield i

    rpath = tmp_path / 'res.org'
    spath = tmp_path / 'state.json'

    def run_view(items, **kwargs):
        TestView(items)._run(
            to=rpath,
            state_path=spath,
            **kwargs,
        )

    def get_state():
        return set(json.loads(spath.read_text()).keys())

    items = []
    run_view([], init=True)
    assert rpath.read_text() == """
# autogenerated!
""".lstrip()
    # TODO do we need to touch state too??

    items.append(
        ('first', OrgNode(heading='i am first')),
    )
    run_view(items)
    assert rpath.read_text() == """
# autogenerated!

* i am first""".lstrip()
    assert get_state() == {'first'}


    items.append(
        ('second', OrgNode('i am second')),
    )
    run_view(items)
    assert rpath.read_text() == """
# autogenerated!

* i am first
* i am second""".lstrip()
    assert get_state() == {'first', 'second'}


    rpath_time = rpath.stat().st_mtime
    spath_time = spath.stat().st_mtime
    run_view(items)
    assert rpath.stat().st_mtime == rpath_time
    assert spath.stat().st_mtime == spath_time




#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading

import my.twitter.all as twi


class TwitterLikesView(StaticView):
    def get_items(self):
        for tweet in twi.likes():
            # likes don't have timestamps (at least from GDPR export data)
            # TODO support it and handle None
            yield node(
                link(url=tweet.permalink, title='liked'),
                body=tweet.text
            )

if __name__ == '__main__':
    TwitterLikesView.main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link, OrgNode
from orger.common import dt_heading

from my.books.kobo import get_books_with_highlights, Highlight


class KoboView(Mirror):
    def get_items(self) -> Mirror.Results:
        def render_highlight(h: Highlight) -> OrgNode:
            # TODO FIXME could use bookmark page??
            heading = 'bookmark' if h.kind == 'bookmark' else (h.text or '')
            body = h.annotation # TODO check if empty
            return node(
                heading=dt_heading(h.dt, heading),
                body=body,
            )

        for page in get_books_with_highlights():
            yield str(page.book), node(
                heading=dt_heading(page.dt, str(page.book)),
                children=[render_highlight(h) for h in page.highlights],
            )


# TODO maybe fixture instead?
test = KoboView.make_test(
    heading='Unsong',
    contains='Singer',
)


if __name__ == '__main__':
    KoboView.main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading, error

import my.hypothesis as hypothesis


class HypView(Mirror):
    def get_items(self) -> Mirror.Results:
        for page in hypothesis.pages():
            # TODO would be nice to signal error upwards? Maybe just yield Exception, rener it in Orger and allow setting error status?
            if isinstance(page, Exception):
                yield error(page)
                continue
            yield node(
                heading=dt_heading(page.created, link(title=page.title, url=page.url)),
                children=[node(
                    heading=dt_heading(hl.created, link(title='context', url=hl.hyp_link)),
                    tags=hl.tags,
                    body=hl.highlight,
                    children=[] if hl.annotation is None else [node(
                        heading=hl.annotation,
                    )]
                ) for hl in page.highlights]
            )


# TODO tests for determinism
# TODO they could also be extracted to common routine and used in each provider
# TODO need to group by source??

if __name__ == '__main__':
    HypView.main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link, timestamp, OrgNode
from orger.common import dt_heading

import datetime
from typing import Iterable, Any

import my.twitter.all as twi

today = datetime.datetime.now()

# TODO FIXME expose Tweet type in twitter.common module?
Tweet = Any

class TwitterView(Mirror):
    @property
    def mode(self) -> str:
        assert self.cmdline_args is not None
        return self.cmdline_args.mode

    def _get_tweets(self) -> Iterable[Tweet]:
        if self.mode == 'all':
            return twi.tweets()
        else:
            # not gonna work on 29 feb!!
            same_day = lambda d: d.day == today.day and d.month == today.month
            return (t for t in twi.tweets() if same_day(t.created_at))

    def _render(self, t: Tweet) -> OrgNode:
        dtime = t.created_at
        text = t.text
        url = t.permalink
        if self.mode == 'all':
            return node(
                heading=dt_heading(dtime, link(title=text, url=url)),
                tags=['tweet'],
            )
        else:
            dd = dtime.replace(year=today.year)
            text = '  ' + text if text.startswith('@') else text # pad replies a bit
            return node(
                heading=timestamp(dd.date(), active=True) + ' ' + f"{link(title='TW', url=url)} at {timestamp(dtime, inactive=True)} {text}",
                tags=['ttweet'],
            )


    def get_items(self) -> Mirror.Results:
        for tweet in sorted(self._get_tweets(), key=lambda t: t.created_at):
            yield self._render(tweet)


def setup_parser(p):
    p.add_argument('--mode', choices=['thatday', 'all'], required=True)


if __name__ == '__main__':
    TwitterView.main(setup_parser=setup_parser)




#!/usr/bin/env python3
"""
Example output:


* [2018-11-28 Wed 22:04] [Baez] Lectures on Classical Mechanics.pdf 12nwnBFvng-[Baez] Lectures on Classical Mechanics.pdf
...
* [2019-01-20 Sun 14:00] [Woit] Quantum Theory, Groups and Representations.pdf 12LWfRf8Gf-[Woit] Quantum Theory, Groups and Representations.pdf
** [2019-01-20 Sun 14:34]  An important corollary of Schurs lemma is the following characterization of irreducible representations of G when G is commutative.
** [2019-01-20 Sun 14:40]  As such it has a differential, which  is  a  linear  map  from  the  tangent  space  at  the  identity  of U (1)  (which here is i R ) to the tangent space at the identity of GL ( n, C ) (which is the space M ( n, C ) of n by n complex matrices).  Th e tangent space at the identity of a Lie group is called a Lie algebra
** [2019-01-20 Sun 14:47]  n general [ Q,H ] 6 = 0, with Q then gen- erating a unitary action on H that does not commute with time evolution and does not imply a conservation law
** [2019-01-20 Sun 14:49]   In the limit N one can make contact with classical notions of spinning objects and angular momentum, but the spin
** [2019-01-20 Sun 15:25]   For further reading
*** [2019-01-20 Sun 15:25] Right. so here I think I missed the bit where he only restricted to spin 1/2
"""


from orger import Mirror
from orger.inorganic import node, link, OrgNode
from orger.common import dt_heading
from orger import pandoc


class PolarView(Mirror):
    def get_items(self):
        from my.reading import polar

        def make_comment(c: polar.Comment) -> OrgNode:
            text = pandoc.to_org(data=c.text, from_='html', logger=self.logger)
            return node(
                heading=dt_heading(c.created, text.splitlines()[0]),
                body=text,
            )

        def make_item(res: polar.Result):
            if isinstance(res, polar.Error):
                # TODO could create error heading from exception automatically? take first line as heading and rest + traceback as the body
                return node(heading='ERROR ' + str(res)) # TODO priority A?
            else:
                book = res
                return node(
                    heading=dt_heading(
                        book.created,
                        # TODO apparently file: is not necessary if the path is absolute?
                        link(url=str(book.path), title=book.title),
                    ),
                    tags=book.tags,
                    children=[node(
                        heading=dt_heading(hl.created, hl.selection),
                        tags=hl.tags,
                        properties=None if hl.color is None else {'POLAR_COLOR': hex2name(hl.color)},
                        children=[make_comment(c) for c in hl.comments],
                    ) for hl in book.items]
                )
        for res in polar.get_entries():
            yield make_item(res)

# not sure about this.. r.g. if the users define their own colors in the future
def hex2name(hexc: str) -> str:
    m = {
        '#ff6900': 'orange',
        '#9900ef': 'violet',
    }
    return m.get(hexc.lower(), hexc)

test = PolarView.make_test(
    heading='I missed the bit where he only restricted to spin'
)



if __name__ == '__main__':
    PolarView.main()




#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading
import traceback

from my.vk import favorites, Favorite

class VkFavs(StaticView):
    def get_items(self):
        for f in favorites():
            if isinstance(f, Favorite):
                yield node(
                    heading=dt_heading(
                        f.dt,
                        f.title if f.url is None else link(url=f.url, title=f.title),
                    ),
                    body=f.text,
                )
            else: # error
                yield node(
                    heading=f'ERROR!',
                    body='\n'.join(traceback.format_exception(Exception, f, None)),
                )


if __name__ == '__main__':
    VkFavs.main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading, error
from orger import pandoc

import my.coding.github as gh
# todo use later: import my.github.ghexport as gh. also careful about using events() -- need to sort?
# I guess makes sense to generally expose get_ methods?


class Github(Mirror):
    def get_items(self) -> Mirror.Results:
        for e in gh.get_events():
            if isinstance(e, Exception):
                yield error(e)
                continue
            # TODO filter only events that have body? e.g. not sure if much point emitting pull requests here
            summary = e.summary
            body = e.body
            if body is None:
                lines = summary.splitlines(keepends=True)
                if len(lines) > 1:
                    summary = lines[0].strip()
                body = ''.join(lines[1:]) # todo meh. hacky, better to extract bodies in the provider properly
            if body.strip() == '':
                body = None

            yield node(
                dt_heading(
                    e.dt,
                    link(url=e.link, title=summary) if e.link is not None else summary
                ),
                body=None if body is None else pandoc.to_org(body, from_='gfm'), # github flavored markdown
            )


if __name__ == '__main__':
    Github.main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading

# TODO add tags?
from my.pocket import articles


class PocketMirror(Mirror):
    def get_items(self) -> Mirror.Results:
        for a in articles():
            yield node(
                heading=dt_heading(
                    a.added,
                    link(title='pocket', url=a.pocket_link)  + ' · ' + link(title=a.title, url=a.url)
                ),
                children=[node(
                    heading=dt_heading(hl.created, hl.text)
                ) for hl in a.highlights]
            )


if __name__ == '__main__':
    PocketMirror.main()



import json
from pathlib import Path
from typing import List, Union, Dict
import logging
import os
import sys
import warnings

PathIsh = Union[str, Path]

from atomicwrites import atomic_write

# TODO hmm. state should be ordered ideally? so it's easy to add/remove items?
# would require storing as list of lists? or use that https://stackoverflow.com/a/6921760/706389
class JsonState:
    def __init__(
            self,
            path: PathIsh,
            dry_run=False,
            default=None,
            logger=logging.getLogger('json-state'),
    ) -> None:
        self.path = Path(path)
        self.dry_run = dry_run

        if default is None:
            default = {}
        self.default = default

        self.state = None
        self.logger = logger
        # TODO for simplicity, write empty if file doesn't exist??

    def __contains__(self, key: str) -> bool:
        return key in self.get()

    def __setitem__(self, key: str, value) -> None:
        current = self.get()
        assert key not in current # just in case
        current[key] = value

        if self.dry_run:
            self.logger.debug('dry run! ignoring %s: %s', key, value)
            return

        with atomic_write(str(self.path), overwrite=True) as fo:
            json.dump(current, fo, indent=1, sort_keys=True)

    def get(self):
        if self.state is None:
            if not self.path.exists():
                self.state = self.default
            else:
                with self.path.open('r') as fo:
                    self.state = json.load(fo)
        return self.state

    def feed(self, key, value, action) -> None:
        if key in self:
            self.logger.debug(f'already handled: %s: %s', key, value)
            return
        self.logger.info(f'adding %s: %s', key, value)
        # TODO not sure about print...
        print(f'adding new item {key}: {value}')
        action()
        self[key] = repr(value)


def test_state(tmp_path):
    import pytest # type: ignore
    path = tmp_path / 'state.json'
    state = JsonState(path)

    def mtime():
        return path.stat().st_mtime

    assert not path.exists()

    res = []

    def feed(k, v):
        def action():
            res.append(v)
        state.feed(k, v, action=action)

    feed('a', 123)
    assert res == [123]

    m1 = mtime()

    feed('a', 456)
    assert res == [123]

    assert mtime() == m1 # shouldn't touch file at all

    state = JsonState(path)

    assert mtime() == m1 # shouldn't touch either

    feed('b', 'abacaba')

    m2 = mtime()

    assert res == [123, 'abacaba']
    feed('a', None)
    assert res == [123, 'abacaba']

    assert mtime() == m2

    # shouldn't trigger because item is already present
    state.feed('a', 'err', lambda: None.whatever) # type: ignore

    with pytest.raises(AttributeError):
        state.feed('hiii', 'error 2 ', lambda: None.whatever) # type: ignore

    assert mtime() == m2 # shouldn't corrupt or modify the file

    state = JsonState(path, dry_run=True)
    feed('c', 'c')
    feed('x', 'y')
    feed('a', 'whatever')

    assert res == [123, 'abacaba', 'c', 'y']
    assert mtime() == m2 # should't modify state in dry run mode





# see https://github.com/karlicoss/pymplate for up-to-date reference


from setuptools import setup, find_packages # type: ignore


def main():
    pkgs = find_packages('src')
    [pkg] = pkgs
    setup(
        name=pkg,
        use_scm_version={
            'version_scheme': 'python-simplified-semver',
            'local_scheme': 'dirty-tag',
        },
        setup_requires=['setuptools_scm'],

        zip_safe=False,

        packages=[pkg, 'orger.modules'],
        # TODO ugh, it's weird. If it's editable install, python3 -m 'orger.modules.module' doesn't work. But it does if installed without editable??
        # for not solved with a symlink...
        package_dir={
            '': 'src',
            'orger.modules': 'modules',
        },
        package_data={pkg: ['py.typed']},

        ## ^^^ this should be mostly automatic and not requiring any changes

        url='https://github.com/karlicoss/orger',
        author='Dima Gerasimov',
        author_email='karlicoss@gmail.com',
        description='Converts data into org-mode',

        install_requires=[
            'appdirs'     , # to keep state files
            'atomicwrites', # to safely append data to a file
        ],
        extras_require={
            'testing': ['pytest'],
            'linting': [
                'pytest',
                'mypy', 'lxml', # lxml for cov report
            ],
        },
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading

from my.media.youtube import watched

from itertools import groupby


class YoutubeView(Mirror):
    def get_items(self) -> Mirror.Results:
        by_url  = lambda w: w.url
        by_when = lambda w: w.when
        items = [
            max(group, key=by_when)
            for _, group in groupby(sorted(watched(), key=by_url), key=by_url)
        ]
        items = sorted(items, key=by_when)
        # TODO for each url only take latest?
        for item in items:
            deleted = item.url == item.title # todo move to HPI?
            l = link(title=item.title + (' (DELETED)' if deleted else ''), url=item.url)
            yield (item.url, node(
                heading=dt_heading(item.when, l),
            ))


if __name__ == '__main__':
    YoutubeView.main()



#!/usr/bin/env python3
"""
Better interface for reading saved reddit posts/comments
"""
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading

from my.reddit import saved


class RedditView(Mirror):
    def get_items(self):
        for s in saved():
            yield s.sid, node(
                # need to make heading lazy due to is_alive
                # eh, can't guess the type of lambda??
                heading=lambda s=s: dt_heading( # type: ignore[misc]
                    s.created,
                    ('[#A] *DEAD*' if self.is_dead_url(s.url) else '') + link(title=s.title, url=s.url) + f' /r/{s.subreddit}'
                ),
                body=s.text,
            )

    # todo this could be generic, i.e. checking all urls?
    def is_dead_url(self, url: str) -> bool:
        assert self.cmdline_args is not None
        if not self.cmdline_args.mark_dead:
            return False
        from kython.knetwork import is_alive
        return is_alive(url)
        # TODO should somehow track handle DELETED comments...
        # sometimes it's also [removed]
        # TODO maybe, track that in reddit provider? since we have all version of saved items over time


def setup_parser(p):
    p.add_argument(
        '--mark-dead',
        action='store_true',
        help="Mark deleted/unavailable content so you could process it ASAP. Mostly useful on first run",
    )

if __name__ == '__main__':
    RedditView.main(setup_parser=setup_parser)



# TODO actually might be useful to store forever a read only version...
# eh, turned out easier to make it lazy..
# def get_saves_with_state():
#     # TODO don't query for dead so often. We actually only want to get it for new items..
#     saves = get_saves(all_=False)
#     with Pool() as p:
#         res = p.map(_helper, saves)
#     dead = [r for r, s in res if not s]
#     # if len(dead) / (len(saves) + 1) > 0.3:
#     #     raise RuntimeError('something must be wrong with reddit! bailing')
#     return res




#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading

import my.rss.all as rss


class RssSubscriptions(StaticView):
    def get_items(self):
        for s in rss.subscriptions():
            yield node(
                link(url=s.url, title=s.title) + ('' if s.subscribed else ' UNSUBSCRIBED'),
            )


if __name__ == '__main__':
    RssSubscriptions.main()



#!/usr/bin/env python3
from orger import StaticView
from orger.inorganic import node, link
from orger.common import dt_heading

import my.stackexchange as se


class Stackexchange(StaticView):
    def get_items(self):
        # TODO adapt for other stackexchange items
        se_data = se.get_data()
        for q in se_data.questions:
            # TODO could emit items along with level, then it would look a bit more natural
            yield node(
                dt_heading(q.creation_date, link(url=q.link, title=q.title)),
                tags=q.tags,
                body=q.body_markdown, # TODO eh, would be useful to have md2org perhaps?
            )


if __name__ == '__main__':
    Stackexchange.main()



#!/usr/bin/env python3
"""
Read-only reddit mirror of comments, submissions and upvoted posts; everything except saved
"""
from orger import Mirror
from orger.inorganic import node, link
from orger.common import dt_heading


from my.reddit import upvoted, submissions, comments


# TODO convert bodies from markdown to org-mode
class RedditAllView(Mirror):
    def get_items(self) -> Mirror.Results:
        yield node(
            'Submissions',
            children=[node( # TODO can also be iterable?
                dt_heading(s.created, link(title=s.title, url=s.url)),
                body=s.text,
            ) for s in submissions()]
        )
        yield node(
            'Comments', # TODO parent thread??
            children=[node(
                dt_heading(c.created, link(title=c.url, url=c.url)),
                body=c.text,
            ) for c in comments()],
        )
        yield node(
            'Upvoted',
            children=[node(
                dt_heading(u.created, link(title=u.title, url=u.url)),
                body=u.text,
             ) for u in upvoted()]
        )


if __name__ == '__main__':
    RedditAllView.main()
